@@ -1,13 +1,23 @@
 import logging
+import queue
 import socket
+import ssl
+from multiprocessing import Queue
 
 import paho.mqtt.client as mqtt
 from paho.mqtt import MQTTException
 
-# global objects
+args = None
 log = None
 mqttc = None
-initial_connection_made = False
+mqtt_initial_connection_made = False
+rx_queue = Queue(maxsize=1000)
+
+
+class MqttMsg:
+    def __init__(self, topic, payload):
+        self.topic = topic
+        self.payload = payload
 
 
 # Callbacks
@@ -16,7 +26,7 @@ def _on_connect(client, userdata, flags, rc, properties=None):
         log.info("Connection successful")
         # Subscribing in on_connect() means that if we lose the connection and
         # reconnect then subscriptions will be renewed.
-        # client.subscribe(subscribe_topic)
+        client.subscribe(f"{args.mqtt_topic_prefix}+/set")
     elif rc == 1:
         log.warning("Connection refused - incorrect protocol version")
     elif rc == 2:
@@ -35,57 +45,65 @@ def _on_disconnect(client, userdata, rc):
     log.info(f"disconnected. rc={rc}.")
 
 
-def _on_message(client, obj, msg):
+def _on_message(client, userdata, msg):
     if msg.retain == 0:
         log.info(f"Receive message ({msg.topic}): {msg.payload}")
     else:
         log.info(f"Receive retained message ({msg.topic}): {msg.payload}")
+    obj = msg.topic, msg.payload
+    try:
+        rx_queue.put(obj, block=False)
+    except queue.Full:
+        log.warning(f"MQTT receiving queue is full, ignoring new message.")
 
 
-def _on_publish(client, obj, mid):
+def _on_publish(client, userdata, mid):
     log.debug("Sent.")
 
 
-def _on_subscribe(client, obj, mid, granted_qos):
+def _on_subscribe(client, userdata, mid, granted_qos):
     log.debug("Subscribed.")
 
 
-def _on_log(client, obj, level, string):
+def _on_log(client, userdata, level, string):
     log.debug(f"{level} | {string}")
 
 
-def mqttc_setup(args):
+def mqttc_setup(config):
+    global args
+    args = config
     global log
     log = logging.getLogger(__name__)
     try:
+        if args.mqtt_clientid is None:
+            if args.mqtt_qos == 0:
+                clientid = ""
+            else:
+                clientid = socket.gethostname()
+        else:
+            clientid = args.mqtt_clientid
         global mqttc
-        mqttc = mqtt.Client(socket.gethostname(), clean_session=(args.mqtt_qos == 0))
-        # if args.mqtt_use_tls:
-        #     if args.mqtt_tls_version == "tlsv1.2":
-        #         tlsVersion = ssl.PROTOCOL_TLSv1_2
-        #     elif args.mqtt_tls_version == "tlsv1.1":
-        #         tlsVersion = ssl.PROTOCOL_TLSv1_1
-        #     elif args.mqtt_tls_version == "tlsv1":
-        #         tlsVersion = ssl.PROTOCOL_TLSv1
-        #     elif args.mqtt_tls_version is None:
-        #         tlsVersion = None
-        #     else:
-        #         print("Unknown TLS version - ignoring")
-        #         tlsVersion = None
-        #
-        #     if not args.mqtt_insecure:
-        #         cert_required = ssl.CERT_REQUIRED
-        #     else:
-        #         cert_required = ssl.CERT_NONE
-        #
-        #     mqttc.tls_set(ca_certs=args.mqtt_cacerts, certfile=None, keyfile=None, cert_reqs=cert_required,
-        #                    tls_version=tlsVersion)
-        #
-        #     if args.mqtt_insecure:
-        #         mqttc.tls_insecure_set(True)
+        mqttc = mqtt.Client(clientid, clean_session=(args.mqtt_qos == 0))
+        if args.mqtt_use_tls:
+            if args.mqtt_tls_version == "tlsv1.2":
+                tlsVersion = ssl.PROTOCOL_TLSv1_2
+            elif args.mqtt_tls_version == "tlsv1.1":
+                tlsVersion = ssl.PROTOCOL_TLSv1_1
+            elif args.mqtt_tls_version == "tlsv1":
+                tlsVersion = ssl.PROTOCOL_TLSv1
+            else:
+                tlsVersion = ssl.PROTOCOL_TLS
+
+            if not args.mqtt_insecure:
+                cert_required = ssl.CERT_REQUIRED
+            else:
+                cert_required = ssl.CERT_NONE
+
+            mqttc.tls_set(ca_certs=args.mqtt_cacerts, certfile=None, keyfile=None, cert_reqs=cert_required,
+                          tls_version=tlsVersion, ciphers=None)
 
         if args.mqtt_user:
-            mqttc.username_pw_set(args.mqtt_user, args.mqtt_password)
+            mqttc.username_pw_set(args.mqtt_user, args.mqtt_pass)
 
         mqttc.on_message = _on_message
         mqttc.on_connect = _on_connect
@@ -102,8 +120,8 @@ def mqttc_setup(args):
         # start loop - let paho manage connection
         mqttc.loop_start()
 
-        global initial_connection_made
-        initial_connection_made = True
+        global mqtt_initial_connection_made
+        mqtt_initial_connection_made = True
         return True
 
     except Exception as ex:
@@ -113,7 +131,6 @@ def mqttc_setup(args):
 
 
 def mqttc_publish(topic, msg, qos=0, retain=False):
-    global mqttc
     try:
         if not mqttc:
             return
@@ -129,8 +146,15 @@ def mqttc_publish(topic, msg, qos=0, retain=False):
         raise ex
 
 
+def mqttc_receive():
+    try:
+        topic, payload = rx_queue.get(block=False)
+        return topic, payload
+    except queue.Empty:
+        return None, None
+
+
 def mqttc_close():
-    global mqttc
     if mqttc:
         mqttc.loop_stop()
         mqttc.disconnect()
